package de.akquinet.jbosscc.needle;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.annotation.Resource;
import javax.ejb.EJBContext;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.transaction.UserTransaction;

import org.apache.commons.collections.map.CompositeMap;
import org.easymock.EasyMock;
import org.easymock.classextension.EasyMockSupport;
import org.jboss.seam.annotations.In;
import org.junit.After;
import org.junit.Before;

public abstract class AbstractTestcase<T> extends EasyMockSupport {

    private final Class<T> clazz;
    protected final Injector injector = new Injector();
    private final Map<Object, Object> mocks = new HashMap<Object, Object>();
    private final Map<String, Object> resources = new HashMap<String, Object>();

    private MockLog mockLog;
    private UserTransaction mockTransaction;
    private EJBContext mockContext;
    private InMemoryDatabaseUtil databaseUtil;

    private T instance;
    private Injector.InjectionObjectCreator lazyEmCreator;

    public AbstractTestcase(final Class<T> clazz) {
        this.clazz = clazz;
    }

    private T createInstance() throws Exception {
        T instance = clazz.newInstance();

        return instance;
    }

    @Before
    public final void setupInstance() throws Exception {
        instance = createInstance();

        mockLog = injector.injectSeamLogger(instance);

        lazyEmCreator = new Injector.InjectionObjectCreator() {
            @Override
            public Object getObjectToBeInjected() {
                return getDatabaseUtil().getEntityManager();
            }
        };

        injectEjbFields();
        injectResourceFields();
        injectSeamFields();

        injector.injectAllDeclaredFields(instance, PersistenceContext.class, lazyEmCreator);

    }

    private void injectResourceFields() {
        List<Field> resourceFields = injector.getAllResourceInjectedFields(instance);

        for (Field f : resourceFields) {
            Class type = f.getType();
            Resource res = f.getAnnotation(Resource.class);

            if (type == EJBContext.class) {
                mockContext = createNiceMock(EJBContext.class);
                mockTransaction = createNiceMock(UserTransaction.class);

                injector.injectResourceParameter(instance, res.mappedName(), mockContext);

                EasyMock.expect(mockContext.getUserTransaction()).andReturn(mockTransaction).anyTimes();
            } else {
                Object mock = createNiceMock(type);

                resources.put(res.mappedName(), mock);

                injector.injectResourceParameter(instance, res.mappedName(), mock);
            }
        }
    }

    private void injectEjbFields() {
        List<Field> ejbFields = injector.getAllEjbInjectedFields(instance);

        for (Field f : ejbFields) {
            Class type = f.getType();

            Object mock = createNiceMock(type);

            mocks.put(type, mock);

            injector.injectEjb(instance, type, mock);
        }
    }

    private void injectSeamFields() {
        List<Field> seamFields = injector.getAllSeamInjectedFields(instance);

        for (Field f : seamFields) {
            if (f.getType() == EntityManager.class) {
                injector.injectAllDeclaredFields(instance, In.class, lazyEmCreator,
                        new Injector.AnnotationChecker<In>() {
                            @Override
                            public boolean checkAnnotation(final Field field, final In annotation) {
                                return field.getType() == EntityManager.class;
                            }
                        });
            } else {

                Class type = f.getType();

                Object mock = createNiceMock(type);

                mocks.put(type, mock);

                injector.injectSeamIn(instance, f.getName(), mock);
            }
        }
    }

    @After
    public void cleanDatabase() {
        if (databaseUtil != null) {
            databaseUtil.deleteAll();
        }
    }

    protected InMemoryDatabaseUtil getDatabaseUtil() {
        if (databaseUtil == null) {
            databaseUtil = new InMemoryDatabaseUtil();
        }

        return databaseUtil;
    }

    public Injector getInjector() {
        return injector;
    }

    protected final EJBContext getMockContext() {
        return mockContext;
    }

    protected final UserTransaction getMockTransaction() {
        return mockTransaction;
    }

    protected final <X> X getResourceMock(final String mappedName) {
        return (X) resources.get(mappedName);
    }

    protected final <X> X getMock(final Class<X> key) {
        return (X) mocks.get(key);
    }

    protected T getObjectUnderTest() {
        return instance;
    }

    protected void forEachMock(final MockOperation operation) {
        Iterator<Map.Entry> i = new CompositeMap(mocks, resources).entrySet().iterator();

        while (i.hasNext()) {
            Map.Entry e = i.next();

            operation.doWithMock(e.getValue());
        }

        if (mockTransaction != null) {
            operation.doWithMock(mockTransaction);
        }

        if (mockContext != null) {
            operation.doWithMock(mockContext);
        }
    }

    protected interface MockOperation {
        void doWithMock(Object mock);
    }

    protected final <X> X resetToDefault(final Object mock) {
        EasyMock.resetToDefault(mock);
        return (X) mock;
    }

    protected final <X> X resetToStrict(final Object mock) {
        EasyMock.resetToStrict(mock);
        return (X) mock;
    }

    protected void setLogVerbose(final boolean verbose) {
        if (mockLog != null) {
            mockLog.setVerbose(verbose);
        }
    }
}
