package de.akquinet.jbosscc.needle;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.ListIterator;
import java.util.ResourceBundle;
import java.util.Set;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;

import org.hibernate.Session;
import org.hibernate.validator.InvalidStateException;
import org.hibernate.validator.InvalidValue;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Util class to manage in-memory-database.
 */
public class InMemoryDatabaseUtil {
    private static final ResourceBundle bundle = ResourceBundle.getBundle("needle");
    private static final String JDBC_URL = bundle.getString("jdbc.url");
    private static final Logger LOG = LoggerFactory.getLogger(InMemoryDatabaseUtil.class);

    /**
     * Must be static in order to save startup time for the tests. However, this
     * requires that deleteAll() is called properly in the setup method.
     */
    private static final EntityManagerFactory ENTITYMANAGER_FACTORY = Persistence
            .createEntityManagerFactory("TestDataModel");

    private final EntityManager entityManager = ENTITYMANAGER_FACTORY.createEntityManager();

    /**
     * Saves the given object in the database.
     *
     * @param <T>
     *            -- type of given object obj
     * @param obj
     *            -- object to save
     * @return -- saved object
     * @throws Exception
     *             -- save objects failed
     */
    public final <T> T saveObject(final T obj) throws Exception {
        return executeInTransaction(new Runnable<T>() {
            public T run(final EntityManager entityManager) {
                entityManager.persist(obj);
                return obj;
            }
        });
    }

    /**
     * Merges the given object with the persistence context.
     *
     * @param <T>
     *            -- type of given object obj
     * @param obj
     *            -- object to merge
     * @return -- merged object
     * @throws Exception
     *             -- merging object failed
     */
    public final <T> T updateObject(final T obj) throws Exception {
        return executeInTransaction(new Runnable<T>() {
            public T run(final EntityManager entityManager) {
                return entityManager.merge(obj);
            }
        });
    }

    /**
     * Finds and returns object of the given id in the persistence context.
     *
     * @param <T>
     *            -- type of searched object
     * @param clazz
     *            -- type of searched object
     * @param id
     *            -- technical id of searched object
     * @return -- found object
     * @throws Exception
     *             -- finding object failed
     */
    public final <T> T loadObject(final Class<T> clazz, final Object id) throws Exception {
        return executeInTransaction(new Runnable<T>() {
            public T run(final EntityManager entityManager) {
                return entityManager.find(clazz, id);
            }
        });
    }

    /**
     * Returns all objects of the given class in persistence context.
     *
     * @param <T>
     *            -- type of searched objects
     * @param clazz
     *            -- type of searched objects
     * @return -- list of found objects
     * @throws Exception
     *             -- finding objects failed
     */
    public final <T> List<T> loadAllObjects(final Class<T> clazz) throws Exception {
        return executeInTransaction(new Runnable<List<T>>() {
            @SuppressWarnings("unchecked")
            public List<T> run(final EntityManager entityManager) {
                return entityManager.createQuery("FROM " + clazz.getSimpleName()).getResultList();
            }
        });
    }

    /**
     * Encapsulates execution of runnable.run() in transactions.
     *
     * @param <T>
     *            -- result type of runnable.run()
     * @param runnable
     *            -- algorithm to execute
     * @param clearAfterCommit
     *            -- <tt>true</tt> triggers entityManager.clear() after
     *            transaction commit
     * @return -- return value of runnable.run()
     * @throws Exception
     *             -- execution failed
     */
    public final <T> T executeInTransaction(final Runnable<T> runnable, final boolean clearAfterCommit)
            throws Exception {
        T result = null;

        try {
            entityManager.getTransaction().begin();
            result = runnable.run(entityManager);
            entityManager.flush();
            entityManager.getTransaction().commit();

            if (clearAfterCommit) {
                entityManager.clear();
            }
        } catch (final InvalidStateException e) {
            final InvalidValue[] invalidValues = e.getInvalidValues();

            for (final InvalidValue value : invalidValues) {
                LOG.warn("Invalid: " + value.getBean() + ":" + value.getMessage() + ":" + value.getPropertyName() + ":"
                        + value.getValue());
            }

            throw e;

        } catch (final Exception exc) {
            exc.printStackTrace();
            LOG.error("While in transaction: ", exc);
            entityManager.getTransaction().rollback();
            throw exc;
        }

        return result;
    }

    /**
     * see executeInTransaction(runnable, clearAfterCommit) .
     *
     * @param <T>
     *            -- result type of runnable.run()
     * @param runnable
     *            -- algorithm to execute
     * @return -- return value of runnable.run()
     * @throws Exception
     *             -- execution failed
     */
    public final <T> T executeInTransaction(final Runnable<T> runnable) throws Exception {
        return executeInTransaction(runnable, true);
    }

    /**
     * Delete everything from the DB: This cannot be done with the JPA, because
     * the order of deletion matters. Instead we directly use a JDBC connection.
     */
    public final void deleteAll() {
        Connection connection = null;

        try {
            connection = DriverManager.getConnection(JDBC_URL);
            connection.setAutoCommit(false);
            final Set<String> tables = getTableNames(connection);
            connection.prepareStatement("set referential_integrity FALSE").execute();
            deleteContent(connection, tables);
            connection.prepareStatement("set referential_integrity TRUE").execute();
        } catch (final Exception exc) {
            LOG.warn("deleteAll caught exception", exc);
        } finally {
            if (connection != null) {
                try {
                    connection.commit();
                    connection.close();
                } catch (final SQLException exc) {
                    LOG.warn("deleteAll.commit", exc);
                }
            }
        }
    }

    private void deleteContent(final Connection connection, final Set<String> tables) throws SQLException {
        final Statement statement = connection.createStatement();

        final ArrayList<String> tempTables = new ArrayList<String>(tables);

        // Loop until all data is deleted: we don't know the correct DROP order,
        // so we have to retry upon failure
        while (!tempTables.isEmpty()) {
            for (final ListIterator<String> iterator = tempTables.listIterator(); iterator.hasNext();) {
                final String table = iterator.next();

                try {
                    statement.executeUpdate("DELETE FROM " + table);
                    iterator.remove();
                } catch (final SQLException exc) {
                    LOG.warn("Ignored exception: " + exc.getMessage() + ". WILL RETRY.");
                }
            }
        }

        statement.close();
    }

    private HashSet<String> getTableNames(final Connection connection) throws SQLException {
        final HashSet<String> tables = new HashSet<String>();

        final ResultSet resultSet = connection.getMetaData().getTables(null, null, "%", new String[] { "TABLE" });

        while (resultSet.next()) {
            tables.add(resultSet.getString("TABLE_NAME"));
        }

        resultSet.close();
        return tables;
    }

    /**
     * Interface for passing algorithms to executeInTransaction().
     *
     * @param <T>
     *            - return type of run()
     */
    public static interface Runnable<T> {
        /**
         * Operation to be automatically called inside executeInTransaction().
         *
         * @param entityManager
         *            -- an entity manager
         * @return -- return value
         * @throws Exception
         *             -- thrown when something failed
         */
        T run(EntityManager entityManager) throws Exception;
    }

    /**
     * Default implementation of Runnable doing nothing.
     */
    public abstract static class VoidRunnable implements Runnable<Object> {
        /**
         * {@inheritDoc}
         */
        public final Object run(final EntityManager entityManager) throws Exception {
            doRun(entityManager);
            return null;
        }

        /**
         * Hook method inside run().
         *
         * @param entityManager
         *            entityManager
         *
         * @throws Exception
         *             -- when something failed
         */
        public abstract void doRun(EntityManager entityManager) throws Exception;
    }

    public final EntityManager getEntityManager() {
        return entityManager;
    }

    public final Session getHibernateSession() {
        return Session.class.cast(entityManager.getDelegate());
    }
}
